"""The main process function."""

from multiprocessing import Pool
from typing import Iterator

import pandas as pd
from sklearn.preprocessing import MinMaxScaler  # type: ignore
from timeseriesfeatures.feature import Feature  # type: ignore

from .distance_correlation_process import distance_correlation_process
from .hsic_process import hsic_process
from .kendall_process import kendall_process
from .mutual_information_process import mutual_information_process
from .pearson_process import pearson_process
from .spearman_process import spearman_process


def process(
    df: pd.DataFrame,
    predictands: list[str],
    max_window: int,
    max_process_features: int = 10,
) -> Iterator[list[Feature]]:
    """Process the dataframe for tsuniverse features."""
    new_df = df.copy()
    for column in df.columns:
        scaler = MinMaxScaler()
        new_df[column] = scaler.fit_transform(new_df[[column]])
    with Pool() as p:
        for predictand in predictands:
            for sub_process in [
                pearson_process,
                mutual_information_process,
                spearman_process,
                kendall_process,
                distance_correlation_process,
                hsic_process,
            ]:
                features = list(sub_process(new_df, predictand, max_window, p))
                features = sorted(
                    features,
                    key=lambda x: abs(x["rank_value"] if "rank_value" in x else 0.0),
                    reverse=True,
                )[:max_process_features]
                yield features



"""The HSIC process function."""

# pylint: disable=duplicate-code,too-many-locals,invalid-name
from typing import Any, Iterator

import pandas as pd
from pyHSICLasso import HSICLasso  # type: ignore
from timeseriesfeatures.feature import FEATURE_TYPE_LAG  # type: ignore
from timeseriesfeatures.feature import Feature  # type: ignore
from timeseriesfeatures.transform import Transform  # type: ignore
from timeseriesfeatures.transforms import TRANSFORMS  # type: ignore

_HSIC_CACHE: dict[str, Feature] = {}


def _construct_lagged_features(universe_df: pd.DataFrame, max_lag: int):
    """
    Returns a new DataFrame where each column is a lagged version of each input column.
    """
    lagged_features = {}
    for col in universe_df.columns:
        for lag in range(1, max_lag + 1):
            lagged_features[f"{col}_lag{lag}"] = universe_df[col].shift(lag)
    return pd.DataFrame(lagged_features, index=universe_df.index).fillna(0.0)


def _hsic_lag_selector(
    universe_df: pd.DataFrame, target_series: pd.Series, max_lag: int, top_k=10
):
    # Step 1: Create lagged feature DataFrame
    lagged_df = _construct_lagged_features(universe_df, max_lag).dropna()

    # Step 2: Align target and features
    aligned_target = target_series.loc[lagged_df.index]

    # Step 3: Convert to NumPy arrays
    X = lagged_df.values
    y = aligned_target.fillna(0.0).values.reshape(-1, 1).flatten()  # type: ignore

    # Step 4: Run HSIC Lasso
    hsic_lasso = HSICLasso()
    hsic_lasso.input(X, y, lagged_df.columns.tolist())
    hsic_lasso.regression(num_feat=top_k)

    # Step 5: Return selected features and scores
    selected_features = hsic_lasso.get_features()
    selected_scores = hsic_lasso.get_index_score()

    return (
        pd.DataFrame(
            {
                "feature": selected_features,
                "hsic_score": selected_scores,
            }
        )
        .sort_values(by="hsic_score", ascending=False)
        .reset_index(drop=True)
    )


def hsic_positive_lags(
    predictor: pd.Series,
    predictand: pd.Series,
    max_window: int,
    predictor_transform: Transform,
    column: str,
) -> Feature:
    """Calculate the best distance correlation for the 2 series within a lag window"""
    df = _hsic_lag_selector(predictor.to_frame(), predictand, max_window)
    top_feature = df["feature"].tolist()[0]
    top_hsic_score = df["hsic_score"].tolist()[0]
    lag = int(top_feature.split("_")[-1].replace("lag", ""))

    return Feature(
        feature_type=FEATURE_TYPE_LAG,
        columns=[column],
        value1=int(lag),
        transform=str(predictor_transform),
        rank_value=top_hsic_score,
        rank_type="hsic",
    )


def hsic_process(
    df: pd.DataFrame,
    predictand: str,
    max_window: int,
    pool: Any,
) -> Iterator[Feature]:
    """Process the dataframe for distance correlation features."""
    predictors = df.columns.values.tolist()
    cached_predictors = []
    for predictor in predictors:
        for transform in TRANSFORMS:
            key = "_".join(sorted([predictor, transform, predictand]))
            feature = _HSIC_CACHE.get(key)
            if feature is not None:
                yield feature
                cached_predictors.append(predictor)
    for transform in TRANSFORMS:
        for feature in pool.starmap(
            hsic_positive_lags,
            [
                (df[x], df[predictand], max_window, transform, x)
                for x in df.columns.values.tolist()
                if x != predictand and x not in cached_predictors
            ],
        ):
            if feature is None:
                continue
            key = "_".join(
                sorted(
                    [
                        feature["columns"][0],
                        transform,
                        predictand,
                    ]
                )
            )
            _HSIC_CACHE[key] = feature
            yield feature



"""The distance correlation function."""

# pylint: disable=duplicate-code,too-many-locals
from typing import Any, Iterator

import dcor  # type: ignore
import numpy as np
import pandas as pd
from timeseriesfeatures.feature import FEATURE_TYPE_LAG  # type: ignore
from timeseriesfeatures.feature import Feature  # type: ignore
from timeseriesfeatures.transform import Transform  # type: ignore
from timeseriesfeatures.transforms import TRANSFORMS  # type: ignore

_DISTANCE_CORRELATION_CACHE: dict[str, Feature] = {}


def distance_correlation_positive_lags(
    predictor: pd.Series,
    predictand: pd.Series,
    max_window: int,
    predictor_transform: Transform,
    column: str,
) -> Feature:
    """Calculate the best distance correlation for the 2 series within a lag window"""
    predictand = predictand.dropna()
    predictor = TRANSFORMS[predictor_transform](predictor)
    corrs = []
    lags = range(1, max_window + 1)

    for lag in lags:
        shifted = predictor.shift(lag)
        aligned = pd.concat([predictand, shifted], axis=1).dropna()

        if aligned.shape[0] < 5 or aligned.iloc[:, 1].nunique() < 2:
            corrs.append(0.0)
            continue

        x = aligned.iloc[:, 0].values
        y = aligned.iloc[:, 1].values

        dc = dcor.distance_correlation(x, y)  # type: ignore

        corrs.append(dc)

    best_idx = np.argmax(np.abs(corrs))
    best_lag = lags[best_idx]
    best_corr = corrs[best_idx]
    if np.isnan(best_corr):
        best_corr = 0.0

    return Feature(
        feature_type=FEATURE_TYPE_LAG,
        columns=[column],
        value1=int(best_lag),
        transform=str(predictor_transform),
        rank_value=best_corr,
        rank_type="distance_correlation",
    )


def distance_correlation_process(
    df: pd.DataFrame,
    predictand: str,
    max_window: int,
    pool: Any,
) -> Iterator[Feature]:
    """Process the dataframe for distance correlation features."""
    predictors = df.columns.values.tolist()
    cached_predictors = []
    for predictor in predictors:
        for transform in TRANSFORMS:
            key = "_".join(sorted([predictor, transform, predictand]))
            feature = _DISTANCE_CORRELATION_CACHE.get(key)
            if feature is not None:
                yield feature
                cached_predictors.append(predictor)
    for transform in TRANSFORMS:
        for feature in pool.starmap(
            distance_correlation_positive_lags,
            [
                (df[x], df[predictand], max_window, transform, x)
                for x in df.columns.values.tolist()
                if x != predictand and x not in cached_predictors
            ],
        ):
            if feature is None:
                continue
            key = "_".join(
                sorted(
                    [
                        feature["columns"][0],
                        transform,
                        predictand,
                    ]
                )
            )
            _DISTANCE_CORRELATION_CACHE[key] = feature
            yield feature
